#include "TreeClickFix.h"

HINSTANCE	hInstance = NULL;
LPSTR		mData = NULL;
HWND		hApp = NULL;
HWND		hTreeview = NULL;
HWND		hMDI = NULL;
HANDLE		hMap = NULL;
HTREEITEM	hTreeLast = NULL;

// Parameters: (HANDLE hInstDLL, DWORD, void*)
DllStartup()
{
	hInstance = (HINSTANCE)hInstDLL;
	return 1;
}

// Parameters: (LOADINFO *load)
mLoad()
{
	hMap = CreateFileMapping(INVALID_HANDLE_VALUE,0,PAGE_READWRITE,0,4096,"mIRC");
	mData = (LPSTR)MapViewOfFile(hMap,FILE_MAP_ALL_ACCESS,0,0,0);
	// Let's save mIRC's HWND in a global variable
	hApp = load->mHwnd;

	if (hApp) {
		hTreeview = GetTreeviewHwnd(hApp);
		hMDI = GetMDIHwnd(hApp);

		if (!GetProp(hTreeview, "CLICKFIXED")) {
			// Only set this if some MDI window is active
			if (IsWindow((HWND)SendMessage(hMDI, WM_MDIGETACTIVE, 0L, 0L)))
				hTreeLast = TreeView_GetSelection(hTreeview);
			// Let's set mKeep to TRUE so this DLL says loaded
			load->mKeep = TRUE;
		}
	}
}

// Parameters: (int timeout)
mUnload()
{
	if (!timeout) {
		// Clean up
		if (IsWindow(hTreeview)) {
			RemoveProp(hTreeview, "CLICKFIXED");
			UnSubclassWindow(hInstance, hTreeview);
		}
		// Close communications withs mIRC
		UnmapViewOfFile(mData);
		CloseHandle(hMap);
		return 1;
	}
	return 0;
}

BOOL IsGrouping(HTREEITEM hItem)
{
	if ((TreeView_GetChild(hTreeview, hItem)) && (TreeView_GetParent(hTreeview, hItem)))
		return TRUE;
	return FALSE;
}

// If mouse if over an HTREEITEM, this function will return a window's HWND
HTREEITEM MouseOverTreeItem(void)
{
	RECT		rect;
	HTREEITEM	hRoot, hGroup, hItem;

	// Get the current mouse position
	POINT		pt;
	GetCursorPos(&pt);
	GetWindowRect(hTreeview, &rect);
	
	pt.x -= rect.left;
	pt.y -= rect.top;

	// Scan all items in the treeview
	hRoot = TreeView_GetRoot(hTreeview);
	while (hRoot) {
		// if mouse is in rect, return hRoot
		TreeView_GetItemRect(hTreeview, hRoot, &rect, TRUE);
		if ((pt.x >= rect.left) && (pt.x <= rect.right) && (pt.y >= rect.top) && (pt.y <= rect.bottom)) {
			return hRoot;
		}
		// Scan Groups
		hGroup = TreeView_GetChild(hTreeview, hRoot);
		if (TreeView_GetChild(hTreeview, hGroup)) {
			while (hGroup) {
				// Scan remaining Windows
				hItem = TreeView_GetChild(hTreeview, hGroup);
				while (hItem) {
					// if mouse is in rect, return hItem
					TreeView_GetItemRect(hTreeview, hItem, &rect, TRUE);
					if ((pt.x >= rect.left) && (pt.x <= rect.right) && (pt.y >= rect.top) && (pt.y <= rect.bottom)) {
						return hItem;
					}
					hItem = TreeView_GetNextSibling(hTreeview, hItem);
				}
				hGroup = TreeView_GetNextSibling(hTreeview, hGroup);
			}
		}
		else {
			// Scan remaining Windows
			hItem = TreeView_GetChild(hTreeview, hRoot);
			while (hItem) {
				// if mouse is in rect, return hItem
				TreeView_GetItemRect(hTreeview, hItem, &rect, TRUE);
				if ((pt.x >= rect.left) && (pt.x <= rect.right) && (pt.y >= rect.top) && (pt.y <= rect.bottom)) {
					return hItem;
				}
				hItem = TreeView_GetNextSibling(hTreeview, hItem);
			}
		}
		hRoot = TreeView_GetNextSibling(hTreeview, hRoot);
	}
	return NULL;
}

// Counts the number of children an hItem has...
// I can't believe treeview controls don't support this.
UINT GetChildCount(HWND hwnd, HTREEITEM hitem)
{
	UINT		count = 0;
	HTREEITEM	hChild;
	if (!hitem)
		hChild = TreeView_GetRoot(hTreeview);
	else
		hChild = TreeView_GetChild(hwnd, hitem);
	while (1) {
		if (!hChild) break;
		count++;
		hChild = TreeView_GetNextSibling(hwnd, hChild);
	}
	return count;
}

// This timer is used to minimize, it keeps a bug from occuring
VOID CALLBACK TimerProc(HWND hwnd, UINT uMsg, UINT_PTR idEvent, DWORD dwTime)
{
	HWND		hMdiChild = (HWND)SendMessage(hMDI, WM_MDIGETACTIVE, 0L, 0L);

	if (idEvent == TIMER_ID1) {
		SetWindowPos(hwnd, HWND_BOTTOM, 0, 0, 0, 0, SWP_NOSIZE | SWP_NOMOVE | SWP_HIDEWINDOW);
		SendMessage(hMDI, WM_MDINEXT, 0L, 0L);
		KillTimer(hwnd, TIMER_ID1);
		/*
		if (IsZoomed(hwnd))
			SetTimer(hwnd, TIMER_ID2, 5, TimerProc);
		*/
	}
	/*
	if (hMdiChild != hwnd) {
		ShowWindow(hMdiChild, SW_MAXIMIZE);
		//SendMessage(hMdiChild, WM_MDIMAXIMIZE, 0L, 0L);
		KillTimer(hwnd, TIMER_ID2);
	}
	*/
}

// Actual function that minimized or closes the selected window
void ClickFix(HTREEITEM hItem)
{
	int			shiftState = HIWORD(GetKeyState(VK_SHIFT));
	HWND		hwnd = (HWND)SendMessage(hMDI, WM_MDIGETACTIVE, 0L, 0L);
	HTREEITEM	hSelect = TreeView_GetSelection(hTreeview);

	// Prevent Group Selections from going further
	if ((!IsWindow(hwnd)) || (hSelect != hItem))
		return;
	
	// Reset hTreeLast so we can't re-re-select an item
	hTreeLast = NULL;

	// Minimizes a MDI window
	if (!shiftState)
		SetTimer(hwnd, TIMER_ID1, 50, TimerProc);

	// Using this message for closes so, mIRC can handle the close process
	else
		SendMessage(hwnd, TVM_SELECTITEM, (WPARAM)TVGN_CARET, (LPARAM)hItem);


}

// Parameters: (HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
mProc(signalProc)
{
	switch (uMsg) {
		case TVM_SELECTITEM:
		{
			HTREEITEM	hItem = (HTREEITEM)lParam;
			//mCmd("echo -s TVM_SELECTITEM: %ld %ld", wParam, lParam);
			if (IsGrouping(hItem))
				hTreeLast = NULL;
			else
				hTreeLast = hItem;
			break;
		}
		case WM_LBUTTONDOWN:
		{
			// This case handles clicks on treeview items
			HTREEITEM	hItem = MouseOverTreeItem();	// hItem == Null ;on grouping items
			
			LRESULT lRes = NextWndProc(hInstance, hwnd, uMsg, wParam, lParam);
			if ((hItem) && (!IsGrouping(hItem))) {
				//mCmd("echo -s WM_LBUTTONDOWN: %ld %ld", wParam, lParam);
				if (hItem == hTreeLast)
					ClickFix(hItem);
				else
					hTreeLast = hItem;
			}
			return lRes;
		}
		case WM_DESTROY:
			UnSubclassWindow(hInstance, hwnd);
			PostMessage(hwnd, uMsg, wParam, lParam);
			return 0L;
		default:
			break;
	}
	return NextWndProc(hInstance, hwnd, uMsg, wParam, lParam);
}

mFunc(mfClickFix)
{
	// Report the error, "Already loaded"
	if (GetProp(hTreeview, "CLICKFIXED"))
		mReturn(MR_"ERROR: The treebar click fix is already loaded");

	if (IsWindow(hTreeview)) {
		if (!SafeSubclassWindow(hInstance, hTreeview, signalProc))
			mReturn(MR_"ERROR: Problem subclassing the window");
		SetProp(hTreeview, "CLICKFIXED", (HANDLE)TRUE);
		mReturn(MR_"Loaded Successfully");
	}
	mReturn(MR_"Error: No treebar treeview control found");
}

